﻿using System;
using System.Threading;
using ParkingLot_Plain.Interfaces;
using ParkingLot_Plain.Parkoviste;

namespace ParkingLot_Plain.Brana
{
    public class Brana : IZapVyp, IPrijeloOdjelo
    {
        public bool Debug = true;
        private const int MsZaMinutu = 60000;
        private const int Auto = 0;
        private const int Bus = 1;
        private const int Tir = 2;
        /* souèet 1, ppsti typu - auto, bus, tir */
        private static readonly double[] PpstiVozidel = {0.4, 0.3, 0.3};
        private readonly int _frekvence;
        private readonly IPrijezdyOdjezdy _parkoviste;
        private bool _beziSimulace;
        private int _odjelo;
        private int _prijelo;
        public bool ParkovistePlne;

        private readonly object _syncLock = new object();

        public Brana(IPrijezdyOdjezdy parkoviste, IKonfigurace konfigurace)
        {
            Subscribe(ParkovistePlneEvent.Instance);
            _parkoviste = parkoviste;
            _frekvence = konfigurace.GetFrekvence();
            _prijelo = 0;
            _odjelo = 0;
        }

        public int GetPrijelo()
        {
            return _prijelo;
        }

        public int GetOdjelo()
        {
            return _odjelo;
        }

        public void Vyp()
        {
            _beziSimulace = false;
            Console.Out.WriteLine("Brana vypnuta...");
        }

        public void Zap()
        {
            _beziSimulace = true;

            if (Debug)
            {
                //NOTE: for debuging just run Run();
                Run();
            }
            else
            {
                var branaVlakno = new Thread(Run);
                branaVlakno.Start();
            }
            Console.Out.WriteLine("Brana zapnuta...");
        }



        public void Subscribe(ParkovistePlneEvent parkovistePlneEvent)
        {
            parkovistePlneEvent.Event += ParkovistePlneEventHandling;
        }

        public void ParkovistePlneEventHandling(Object source)
        {
            SetParkovistePlne();
        }

        private bool IsParkovistePlne()
        {
            lock (_syncLock) { return ParkovistePlne; }
        }

        private void SetParkovistePlne()
        {
            lock (_syncLock) { ParkovistePlne = true; }
            
        }

        private void SetParkovisteVolne()
        {
            lock (_syncLock) { ParkovistePlne = false; }
           
        }

        public void Run()
        {
            while (_beziSimulace)
            {
                try
                {
                    Thread.Sleep((int) Math.Round((Exponential(_frekvence)*MsZaMinutu)));
                }
                catch (Exception e)
                {
                    Console.Out.WriteLine(e);
                }

                if (!_beziSimulace) continue;

                if (JePrijezd())
                {
                    Prijezd(TypVozidla());
                }
                else
                {
                    Odjezd(TypVozidla());
                }
            }
        }

        private void Prijezd(int typ)
        {
            if (IsParkovistePlne())
            {
                return;
            }

            try
            {
                switch (typ)
                {
                    case Auto:
                        Println("Prijezd auto.");
                        _parkoviste.PrijeloAuto();
                        break;
                    case Bus:
                        Println("Prijezd bus.");
                        _parkoviste.PrijelBus();
                        break;
                    case Tir:
                        Println("Prijezd tir.");
                        _parkoviste.PrijelTir();
                        break;
                }

                _prijelo++;
            }
            catch (Exception)
            {
                //ignorujeme
            }
        }

        private void Odjezd(int typ)
        {
            try
            {
                switch (typ)
                {
                    case Auto:
                        Println("Odjezd auto.");
                        _parkoviste.OdjeloAuto();
                        break;
                    case Bus:
                        Println("Odjezd bus.");
                        _parkoviste.OdjelBus();
                        break;
                    case Tir:
                        Println("Odjezd tir.");
                        _parkoviste.OdjelTir();
                        break;
                }

                _odjelo++;
                SetParkovisteVolne();
            }
            catch (Exception)
            {
                //ignorujeme
            }
        }

        public static double Exponential(double lambda)
        {
            return -Math.Log(1 - new Random().NextDouble())/lambda;
        }

        private static bool JePrijezd()
        {
            var rand = new Random();
            return rand.NextDouble() < 0.5;
        }

        private static int TypVozidla()
        {
            double interval = 0;
            var ppst = new Random().NextDouble();

            var vybranyTyp = 0;

            for (var i = 0; i < PpstiVozidel.Length; i++)
            {
                //nastaveni hranice intervalu pro losovani
                interval += PpstiVozidel[i];

                //pokud je generovane cislo z daneho intervalu
                if (!(ppst <= interval)) continue;

                vybranyTyp = i;
                break;
            }

            return vybranyTyp;
        }

        private static void Println(String output)
        {
            Console.Out.WriteLine("(Brana) " + output);
        }
    }
}